package solutions.bitbadger.documents.query

import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import kotlin.jvm.Throws
import solutions.bitbadger.documents.query.byFields as byFieldsBase

/**
 * Functions to count documents
 */
object CountQuery {

    /**
     * Query to count all documents in a table
     *
     * @param tableName The table in which to count documents (may include schema)
     * @return A query to count documents
     */
    @JvmStatic
    fun all(tableName: String) =
        "SELECT COUNT(*) AS it FROM $tableName"

    /**
     * Query to count documents matching the given fields
     *
     * @param tableName The table in which to count documents (may include schema)
     * @param fields The field comparisons for the count
     * @param howMatched How fields should be compared (optional, defaults to ALL)
     * @return A query to count documents matching the given fields
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        byFieldsBase(all(tableName), fields, howMatched)

    /**
     * Query to count documents via JSON containment (PostgreSQL only)
     *
     * @param tableName The table in which to count documents (may include schema)
     * @return A query to count documents via JSON containment
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byContains(tableName: String) =
        statementWhere(all(tableName), Where.jsonContains())

    /**
     * Query to count documents via a JSON path match (PostgreSQL only)
     *
     * @param tableName The table in which to count documents (may include schema)
     * @return A query to count documents via a JSON path match
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String) =
        statementWhere(all(tableName), Where.jsonPathMatches())
}
