package solutions.bitbadger.documents.query

import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import kotlin.jvm.Throws
import solutions.bitbadger.documents.query.byFields as byFieldsBase
import solutions.bitbadger.documents.query.byId as byIdBase

/**
 * Functions to delete documents
 */
object DeleteQuery {

    /**
     * Query to delete documents from a table
     *
     * @param tableName The table in which documents should be deleted (may include schema)
     * @return A query to delete documents
     */
    private fun delete(tableName: String) =
        "DELETE FROM $tableName"

    /**
     * Query to delete a document by its ID
     *
     * @param tableName The table from which documents should be deleted (may include schema)
     * @param docId The ID of the document (optional, used for type checking)
     * @return A query to delete a document by its ID
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TKey> byId(tableName: String, docId: TKey? = null) =
         byIdBase(delete(tableName), docId)

    /**
     * Query to delete documents matching the given fields
     *
     * @param tableName The table from which documents should be deleted (may include schema)
     * @param fields The field comparisons for documents to be deleted
     * @param howMatched How fields should be compared (optional, defaults to ALL)
     * @return A query to delete documents matching for the given fields
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        byFieldsBase(delete(tableName), fields, howMatched)

    /**
     * Query to delete documents via JSON containment (PostgreSQL only)
     *
     * @param tableName The table from which documents should be deleted (may include schema)
     * @return A query to delete documents via JSON containment
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byContains(tableName: String) =
        statementWhere(delete(tableName), Where.jsonContains())

    /**
     * Query to delete documents via a JSON path match (PostgreSQL only)
     *
     * @param tableName The table from which documents should be deleted (may include schema)
     * @return A query to delete documents via a JSON path match
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String) =
        statementWhere(delete(tableName), Where.jsonPathMatches())
}
