package solutions.bitbadger.documents.query

import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import kotlin.jvm.Throws

/**
 * Functions to check for document existence
 */
object ExistsQuery {

    /**
     * Query to check for document existence in a table
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @param where The `WHERE` clause with the existence criteria
     * @return A query to check document existence
     */
    private fun exists(tableName: String, where: String) =
        "SELECT EXISTS (SELECT 1 FROM $tableName WHERE $where) AS it"

    /**
     * Query to check for document existence by ID
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @param docId The ID of the document (optional, used for type checking)
     * @return A query to determine document existence by ID
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TKey> byId(tableName: String, docId: TKey? = null) =
        exists(tableName, Where.byId(docId = docId))

    /**
     * Query to check for document existence matching the given fields
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @param fields The field comparisons for the existence check
     * @param howMatched How fields should be compared (optional, defaults to ALL)
     * @return A query to determine document existence for the given fields
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        exists(tableName, Where.byFields(fields, howMatched))

    /**
     * Query to check for document existence via JSON containment (PostgreSQL only)
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @return A query to determine document existence via JSON containment
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byContains(tableName: String) =
        exists(tableName, Where.jsonContains())

    /**
     * Query to check for document existence via a JSON path match (PostgreSQL only)
     *
     * @param tableName The table in which existence should be checked (may include schema)
     * @return A query to determine document existence via a JSON path match
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String) =
        exists(tableName, Where.jsonPathMatches())
}
