package solutions.bitbadger.documents.query

import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import kotlin.jvm.Throws
import solutions.bitbadger.documents.query.byId as byIdBase
import solutions.bitbadger.documents.query.byFields as byFieldsBase

/**
 * Functions to retrieve documents
 */
object FindQuery {

    /**
     * Query to retrieve all documents from a table
     *
     * @param tableName The table from which documents should be retrieved (may include schema)
     * @return A query to retrieve documents
     */
    @JvmStatic
    fun all(tableName: String) =
        "SELECT data FROM $tableName"

    /**
     * Query to retrieve a document by its ID
     *
     * @param tableName The table from which documents should be retrieved (may include schema)
     * @param docId The ID of the document (optional, used for type checking)
     * @return A query to retrieve a document by its ID
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TKey> byId(tableName: String, docId: TKey? = null) =
        byIdBase(all(tableName), docId)

    /**
     * Query to retrieve documents matching the given fields
     *
     * @param tableName The table from which documents should be retrieved (may include schema)
     * @param fields The field comparisons for matching documents to retrieve
     * @param howMatched How fields should be compared (optional, defaults to ALL)
     * @return A query to retrieve documents matching the given fields
     * @throws DocumentException If the dialect has not been set
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        byFieldsBase(all(tableName), fields, howMatched)

    /**
     * Query to retrieve documents via JSON containment (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved (may include schema)
     * @return A query to retrieve documents via JSON containment
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byContains(tableName: String) =
        statementWhere(all(tableName), Where.jsonContains())

    /**
     * Query to retrieve documents via a JSON path match (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved (may include schema)
     * @return A query to retrieve documents via a JSON path match
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String) =
        statementWhere(all(tableName), Where.jsonPathMatches())
}
