package solutions.bitbadger.documents.query

import solutions.bitbadger.documents.*
import kotlin.jvm.Throws
import solutions.bitbadger.documents.query.byFields as byFieldsBase
import solutions.bitbadger.documents.query.byId as byIdBase

/**
 * Functions to create queries to patch (partially update) JSON documents
 */
object PatchQuery {

    /**
     * Create an `UPDATE` statement to patch documents
     *
     * @param tableName The table to be updated
     * @return A query to patch documents
     */
    private fun patch(tableName: String) =
        when (Configuration.dialect("create patch query")) {
            Dialect.POSTGRESQL -> "data || :data"
            Dialect.SQLITE -> "json_patch(data, json(:data))"
        }.let { "UPDATE $tableName SET data = $it" }

    /**
     * A query to patch (partially update) a JSON document by its ID
     *
     * @param tableName The name of the table where the document is stored
     * @param docId The ID of the document to be updated (optional, used for type checking)
     * @return A query to patch a JSON document by its ID
     * @throws DocumentException If the dialect is not configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun <TKey> byId(tableName: String, docId: TKey? = null) =
        byIdBase(patch(tableName), docId)

    /**
     * A query to patch (partially update) a JSON document using field match criteria
     *
     * @param tableName The name of the table where the documents are stored
     * @param fields The field criteria
     * @param howMatched How the fields should be matched (optional, defaults to `ALL`)
     * @return A query to patch JSON documents by field match criteria
     * @throws DocumentException If the dialect is not configured
     */
    @Throws(DocumentException::class)
    @JvmStatic
    @JvmOverloads
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        byFieldsBase(patch(tableName), fields, howMatched)

    /**
     * A query to patch (partially update) a JSON document by JSON containment (PostgreSQL only)
     *
     * @param tableName The name of the table where the document is stored
     * @return A query to patch JSON documents by JSON containment
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byContains(tableName: String) =
        statementWhere(patch(tableName), Where.jsonContains())

    /**
     * A query to patch (partially update) a JSON document by JSON path match (PostgreSQL only)
     *
     * @param tableName The name of the table where the document is stored
     * @return A query to patch JSON documents by JSON path match
     * @throws DocumentException If the database dialect is not PostgreSQL
     */
    @Throws(DocumentException::class)
    @JvmStatic
    fun byJsonPath(tableName: String) =
        statementWhere(patch(tableName), Where.jsonPathMatches())
}
