package solutions.bitbadger.documents.core.tests.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.AutoId;
import solutions.bitbadger.documents.Configuration;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.query.DocumentQuery;
import solutions.bitbadger.documents.core.tests.ForceDialect;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;

/**
 * Unit tests for the `Document` object
 */
@DisplayName("Core | Java | Query | DocumentQuery")
final public class DocumentQueryTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    public void cleanUp() {
        ForceDialect.none();
    }

    @Test
    @DisplayName("insert generates no auto ID | PostgreSQL")
    public void insertNoAutoPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("INSERT INTO %s VALUES (:data)", TEST_TABLE), DocumentQuery.insert(TEST_TABLE));
    }

    @Test
    @DisplayName("insert generates no auto ID | SQLite")
    public void insertNoAutoSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals(String.format("INSERT INTO %s VALUES (:data)", TEST_TABLE), DocumentQuery.insert(TEST_TABLE));
    }

    @Test
    @DisplayName("insert generates auto number | PostgreSQL")
    public void insertAutoNumberPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("INSERT INTO %1$s VALUES (:data::jsonb || ('{\"id\":' "
                        + "|| (SELECT COALESCE(MAX((data->>'id')::numeric), 0) + 1 FROM %1$s) || '}')::jsonb)",
                        TEST_TABLE),
                DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER));
    }

    @Test
    @DisplayName("insert generates auto number | SQLite")
    public void insertAutoNumberSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals(String.format("INSERT INTO %1$s VALUES (json_set(:data, '$.id', "
                        + "(SELECT coalesce(max(data->>'id'), 0) + 1 FROM %1$s)))", TEST_TABLE),
                DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER));
    }

    @Test
    @DisplayName("insert generates auto UUID | PostgreSQL")
    public void insertAutoUUIDPostgres() throws DocumentException {
        ForceDialect.postgres();
        final String query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID);
        assertTrue(query.startsWith(String.format("INSERT INTO %s VALUES (:data::jsonb || '{\"id\":\"", TEST_TABLE)),
                String.format("Query start not correct (actual: %s)", query));
        assertTrue(query.endsWith("\"}')"), "Query end not correct");
    }

    @Test
    @DisplayName("insert generates auto UUID | SQLite")
    public void insertAutoUUIDSQLite() throws DocumentException {
        ForceDialect.sqlite();
        final String query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID);
        assertTrue(query.startsWith(String.format("INSERT INTO %s VALUES (json_set(:data, '$.id', '", TEST_TABLE)),
                String.format("Query start not correct (actual: %s)", query));
        assertTrue(query.endsWith("'))"), "Query end not correct");
    }

    @Test
    @DisplayName("insert generates auto random string | PostgreSQL")
    public void insertAutoRandomPostgres() throws DocumentException {
        try {
            ForceDialect.postgres();
            Configuration.idStringLength = 8;
            final String query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING);
            final String start = String.format("INSERT INTO %s VALUES (:data::jsonb || '{\"id\":\"", TEST_TABLE);
            final String end = "\"}')";
            assertTrue(query.startsWith(start), String.format("Query start not correct (actual: %s)", query));
            assertTrue(query.endsWith(end), "Query end not correct");
            assertEquals(8, query.replace(start, "").replace(end, "").length(), "Random string length incorrect");
        } finally {
            Configuration.idStringLength = 16;
        }
    }

    @Test
    @DisplayName("insert generates auto random string | SQLite")
    public void insertAutoRandomSQLite() throws DocumentException {
        ForceDialect.sqlite();
        final String query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING);
        final String start = String.format("INSERT INTO %s VALUES (json_set(:data, '$.id', '", TEST_TABLE);
        final String end = "'))";
        assertTrue(query.startsWith(start), String.format("Query start not correct (actual: %s)", query));
        assertTrue(query.endsWith(end), "Query end not correct");
        assertEquals(Configuration.idStringLength, query.replace(start, "").replace(end, "").length(),
                "Random string length incorrect");
    }

    @Test
    @DisplayName("insert fails when no dialect is set")
    public void insertFailsUnknown() {
        assertThrows(DocumentException.class, () -> DocumentQuery.insert(TEST_TABLE));
    }

    @Test
    @DisplayName("save generates correctly")
    public void save() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format(
                        "INSERT INTO %s VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data",
                        TEST_TABLE),
                DocumentQuery.save(TEST_TABLE), "INSERT ON CONFLICT UPDATE statement not constructed correctly");
    }

    @Test
    @DisplayName("update generates successfully")
    public void update() {
        assertEquals(String.format("UPDATE %s SET data = :data", TEST_TABLE), DocumentQuery.update(TEST_TABLE),
                "Update query not constructed correctly");
    }
}
