package solutions.bitbadger.documents.core.tests.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.Parameter;
import solutions.bitbadger.documents.ParameterType;
import solutions.bitbadger.documents.query.RemoveFieldsQuery;
import solutions.bitbadger.documents.core.tests.ForceDialect;

import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;

/**
 * Unit tests for the `RemoveFields` object
 */
@DisplayName("Core | Java | Query | RemoveFieldsQuery")
final public class RemoveFieldsQueryTest {

    /**
     * Reset the dialect
     */
    @AfterEach
    public void cleanUp() {
        ForceDialect.none();
    }

    @Test
    @DisplayName("byId generates correctly | PostgreSQL")
    public void byIdPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("UPDATE %s SET data = data - :name::text[] WHERE data->>'id' = :id", TEST_TABLE),
                RemoveFieldsQuery.byId(TEST_TABLE, List.of(new Parameter<>(":name", ParameterType.STRING, "{a,z}"))),
                "Remove Fields query not constructed correctly");
    }

    @Test
    @DisplayName("byId generates correctly | SQLite")
    public void byIdSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals(String.format("UPDATE %s SET data = json_remove(data, :name0, :name1) WHERE data->>'id' = :id",
                        TEST_TABLE),
                RemoveFieldsQuery.byId(TEST_TABLE, List.of(new Parameter<>(":name0", ParameterType.STRING, "a"),
                        new Parameter<>(":name1", ParameterType.STRING, "z"))),
                "Remove Field query not constructed correctly");
    }

    @Test
    @DisplayName("byFields generates correctly | PostgreSQL")
    public void byFieldsPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("UPDATE %s SET data = data - :name::text[] WHERE data->>'f' > :g", TEST_TABLE),
                RemoveFieldsQuery.byFields(TEST_TABLE, List.of(new Parameter<>(":name", ParameterType.STRING, "{b,c}")),
                        List.of(Field.greater("f", "", ":g"))),
                "Remove Field query not constructed correctly");
    }

    @Test
    @DisplayName("byFields generates correctly | SQLite")
    public void byFieldsSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals(String.format("UPDATE %s SET data = json_remove(data, :name0, :name1) WHERE data->>'f' > :g",
                        TEST_TABLE),
                RemoveFieldsQuery.byFields(TEST_TABLE, List.of(new Parameter<>(":name0", ParameterType.STRING, "b"),
                                new Parameter<>(":name1", ParameterType.STRING, "c")),
                        List.of(Field.greater("f", "", ":g"))),
                "Remove Field query not constructed correctly");
    }

    @Test
    @DisplayName("byContains generates correctly | PostgreSQL")
    public void byContainsPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format("UPDATE %s SET data = data - :name::text[] WHERE data @> :criteria", TEST_TABLE),
                RemoveFieldsQuery.byContains(TEST_TABLE,
                        List.of(new Parameter<>(":name", ParameterType.STRING, "{m,n}"))),
                "Remove Field query not constructed correctly");
    }

    @Test
    @DisplayName("byContains fails | SQLite")
    public void byContainsSQLite() {
        ForceDialect.sqlite();
        assertThrows(DocumentException.class, () -> RemoveFieldsQuery.byContains(TEST_TABLE, List.of()));
    }

    @Test
    @DisplayName("byJsonPath generates correctly | PostgreSQL")
    public void byJsonPathPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals(String.format(
                        "UPDATE %s SET data = data - :name::text[] WHERE jsonb_path_exists(data, :path::jsonpath)",
                        TEST_TABLE),
                RemoveFieldsQuery.byJsonPath(TEST_TABLE,
                        List.of(new Parameter<>(":name", ParameterType.STRING, "{o,p}"))),
                "Remove Field query not constructed correctly");
    }

    @Test
    @DisplayName("byJsonPath fails | SQLite")
    public void byJsonPathSQLite() {
        ForceDialect.sqlite();
        assertThrows(DocumentException.class, () -> RemoveFieldsQuery.byJsonPath(TEST_TABLE, List.of()));
    }
}
