package solutions.bitbadger.documents.core.tests.java;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import solutions.bitbadger.documents.DocumentException;
import solutions.bitbadger.documents.Field;
import solutions.bitbadger.documents.FieldMatch;
import solutions.bitbadger.documents.query.Where;
import solutions.bitbadger.documents.core.tests.ForceDialect;

import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

/**
 * Unit tests for the `Where` object
 */
@DisplayName("Core | Java | Query | Where")
final public class WhereTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    public void cleanUp() {
        ForceDialect.none();
    }

    @Test
    @DisplayName("byFields is blank when given no fields")
    public void byFieldsBlankIfEmpty() throws DocumentException {
        assertEquals("", Where.byFields(List.of()));
    }

    @Test
    @DisplayName("byFields generates one numeric field | PostgreSQL")
    public void byFieldsOneFieldPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("(data->>'it')::numeric = :that", Where.byFields(List.of(Field.equal("it", 9, ":that"))));
    }

    @Test
    @DisplayName("byFields generates one alphanumeric field | PostgreSQL")
    public void byFieldsOneAlphaFieldPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data->>'it' = :that", Where.byFields(List.of(Field.equal("it", "", ":that"))));
    }

    @Test
    @DisplayName("byFields generates one field | SQLite")
    public void byFieldsOneFieldSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("data->>'it' = :that", Where.byFields(List.of(Field.equal("it", "", ":that"))));
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ default match | PostgreSQL")
    public void byFieldsMultipleDefaultPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
                Where.byFields(List.of(
                        Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))));
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ default match | SQLite")
    public void byFieldsMultipleDefaultSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
                Where.byFields(List.of(
                        Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))));
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ ANY match | PostgreSQL")
    public void byFieldsMultipleAnyPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
                Where.byFields(List.of(
                        Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                        FieldMatch.ANY));
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ ANY match | SQLite")
    public void byFieldsMultipleAnySQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
                Where.byFields(List.of(
                        Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                        FieldMatch.ANY));
    }

    @Test
    @DisplayName("byId generates defaults for alphanumeric key | PostgreSQL")
    public void byIdDefaultAlphaPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data->>'id' = :id", Where.byId(":id", ""));
    }

    @Test
    @DisplayName("byId generates defaults for numeric key | PostgreSQL")
    public void byIdDefaultNumericPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("(data->>'id')::numeric = :id", Where.byId(":id", 5));
    }

    @Test
    @DisplayName("byId generates defaults | SQLite")
    public void byIdDefaultSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("data->>'id' = :id", Where.byId(":id", ""));
    }

    @Test
    @DisplayName("byId generates named ID | PostgreSQL")
    public void byIdDefaultNamedPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data->>'id' = :key", Where.byId(":key"));
    }

    @Test
    @DisplayName("byId generates named ID | SQLite")
    public void byIdDefaultNamedSQLite() throws DocumentException {
        ForceDialect.sqlite();
        assertEquals("data->>'id' = :key", Where.byId(":key"));
    }

    @Test
    @DisplayName("jsonContains generates defaults | PostgreSQL")
    public void jsonContainsDefaultPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data @> :criteria", Where.jsonContains());
    }

    @Test
    @DisplayName("jsonContains generates named parameter | PostgreSQL")
    public void jsonContainsNamedPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("data @> :it", Where.jsonContains(":it"));
    }

    @Test
    @DisplayName("jsonContains fails | SQLite")
    public void jsonContainsFailsSQLite() {
        ForceDialect.sqlite();
        assertThrows(DocumentException.class, Where::jsonContains);
    }

    @Test
    @DisplayName("jsonPathMatches generates defaults | PostgreSQL")
    public void jsonPathMatchDefaultPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("jsonb_path_exists(data, :path::jsonpath)", Where.jsonPathMatches());
    }

    @Test
    @DisplayName("jsonPathMatches generates named parameter | PostgreSQL")
    public void jsonPathMatchNamedPostgres() throws DocumentException {
        ForceDialect.postgres();
        assertEquals("jsonb_path_exists(data, :jp::jsonpath)", Where.jsonPathMatches(":jp"));
    }

    @Test
    @DisplayName("jsonPathMatches fails | SQLite")
    public void jsonPathFailsSQLite() {
        ForceDialect.sqlite();
        assertThrows(DocumentException.class, Where::jsonPathMatches);
    }
}
