package solutions.bitbadger.documents.core.tests.java.integration;

import solutions.bitbadger.documents.*;
import solutions.bitbadger.documents.core.tests.integration.ThrowawayDatabase;
import solutions.bitbadger.documents.java.Results;
import solutions.bitbadger.documents.query.CountQuery;
import solutions.bitbadger.documents.query.DeleteQuery;
import solutions.bitbadger.documents.query.FindQuery;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Collection;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;
import static solutions.bitbadger.documents.core.tests.TypesKt.TEST_TABLE;
import static solutions.bitbadger.documents.java.extensions.ConnExt.*;
import static solutions.bitbadger.documents.query.QueryUtils.orderBy;

final public class CustomFunctions {

    public static void listEmpty(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        deleteByFields(db.getConn(), TEST_TABLE, List.of(Field.exists(Configuration.idField)));
        Collection<JsonDocument> result =
                customList(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), JsonDocument.class, Results::fromData);
        assertEquals(0, result.size(), "There should have been no results");
    }

    public static void listAll(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        Collection<JsonDocument> result =
                customList(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), JsonDocument.class, Results::fromData);
        assertEquals(5, result.size(), "There should have been 5 results");
    }

    public static void jsonArrayEmpty(ThrowawayDatabase db) throws DocumentException {
        assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "The test table should be empty");
        assertEquals("[]", customJsonArray(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), Results::jsonFromData),
                "An empty list was not represented correctly");
    }

    public static void jsonArraySingle(ThrowawayDatabase db) throws DocumentException {
        insert(db.getConn(), TEST_TABLE, new ArrayDocument("one", List.of("2", "3")));
        assertEquals(JsonFunctions.maybeJsonB("[{\"id\":\"one\",\"values\":[\"2\",\"3\"]}]"),
                customJsonArray(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), Results::jsonFromData),
                "A single document list was not represented correctly");
    }

    public static void jsonArrayMany(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        assertEquals(JsonFunctions.maybeJsonB("[{\"id\":\"first\",\"values\":[\"a\",\"b\",\"c\"]},"
                        + "{\"id\":\"second\",\"values\":[\"c\",\"d\",\"e\"]},"
                        + "{\"id\":\"third\",\"values\":[\"x\",\"y\",\"z\"]}]"),
                customJsonArray(db.getConn(), FindQuery.all(TEST_TABLE) + orderBy(List.of(Field.named("id"))),
                        List.of(), Results::jsonFromData),
                "A multiple document list was not represented correctly");
    }

    public static void writeJsonArrayEmpty(ThrowawayDatabase db) throws DocumentException {
        assertEquals(0L, countAll(db.getConn(), TEST_TABLE), "The test table should be empty");
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeCustomJsonArray(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), writer, Results::jsonFromData);
        assertEquals("[]", output.toString(), "An empty list was not represented correctly");
    }

    public static void writeJsonArraySingle(ThrowawayDatabase db) throws DocumentException {
        insert(db.getConn(), TEST_TABLE, new ArrayDocument("one", List.of("2", "3")));
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeCustomJsonArray(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), writer, Results::jsonFromData);
        assertEquals(JsonFunctions.maybeJsonB("[{\"id\":\"one\",\"values\":[\"2\",\"3\"]}]"), output.toString(),
                "A single document list was not represented correctly");
    }

    public static void writeJsonArrayMany(ThrowawayDatabase db) throws DocumentException {
        for (ArrayDocument doc : ArrayDocument.testDocuments) { insert(db.getConn(), TEST_TABLE, doc); }
        final StringWriter output = new StringWriter();
        final PrintWriter writer = new PrintWriter(output);
        writeCustomJsonArray(db.getConn(), FindQuery.all(TEST_TABLE) + orderBy(List.of(Field.named("id"))), List.of(),
                writer, Results::jsonFromData);
        assertEquals(JsonFunctions.maybeJsonB("[{\"id\":\"first\",\"values\":[\"a\",\"b\",\"c\"]},"
                        + "{\"id\":\"second\",\"values\":[\"c\",\"d\",\"e\"]},"
                        + "{\"id\":\"third\",\"values\":[\"x\",\"y\",\"z\"]}]"),
                output.toString(), "A multiple document list was not represented correctly");
    }

    public static void singleNone(ThrowawayDatabase db) throws DocumentException {
        assertFalse(
                customSingle(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), JsonDocument.class, Results::fromData)
                        .isPresent(),
                "There should not have been a document returned");
    }

    public static void singleOne(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertTrue(
                customSingle(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), JsonDocument.class, Results::fromData)
                        .isPresent(),
                "There should have been a document returned");
    }

    public static void jsonSingleNone(ThrowawayDatabase db) throws DocumentException {
        assertEquals("{}", customJsonSingle(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), Results::jsonFromData),
                "An empty document was not represented correctly");
    }

    public static void jsonSingleOne(ThrowawayDatabase db) throws DocumentException {
        insert(db.getConn(), TEST_TABLE, new ArrayDocument("me", List.of("myself", "i")));
        assertEquals(JsonFunctions.maybeJsonB("{\"id\":\"me\",\"values\":[\"myself\",\"i\"]}"),
                customJsonSingle(db.getConn(), FindQuery.all(TEST_TABLE), List.of(), Results::jsonFromData),
                "A single document was not represented correctly");
    }

    public static void nonQueryChanges(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(5L,
                customScalar(db.getConn(), CountQuery.all(TEST_TABLE), List.of(), Long.class, Results::toCount),
                "There should have been 5 documents in the table");
        customNonQuery(db.getConn(), String.format("DELETE FROM %s", TEST_TABLE));
        assertEquals(0L,
                customScalar(db.getConn(), CountQuery.all(TEST_TABLE), List.of(), Long.class, Results::toCount),
                "There should have been no documents in the table");
    }

    public static void nonQueryNoChanges(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(5L,
                customScalar(db.getConn(), CountQuery.all(TEST_TABLE), List.of(), Long.class, Results::toCount),
                "There should have been 5 documents in the table");
        customNonQuery(db.getConn(), DeleteQuery.byId(TEST_TABLE, "eighty-two"),
                List.of(new Parameter<>(":id", ParameterType.STRING, "eighty-two")));
        assertEquals(5L,
                customScalar(db.getConn(), CountQuery.all(TEST_TABLE), List.of(), Long.class, Results::toCount),
                "There should still have been 5 documents in the table");
    }

    public static void scalar(ThrowawayDatabase db) throws DocumentException {
        JsonDocument.load(db);
        assertEquals(3L,
                customScalar(db.getConn(), String.format("SELECT 3 AS it FROM %s LIMIT 1", TEST_TABLE), List.of(),
                        Long.class, Results::toCount),
                "The number 3 should have been returned");
    }
}
