package solutions.bitbadger.documents.core.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.query.Where
import kotlin.test.assertEquals

/**
 * Unit tests for the `Where` object
 */
@DisplayName("Core | Kotlin | Query | Where")
class WhereTest {

    /**
     * Clear the connection string (resets Dialect)
     */
    @AfterEach
    fun cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName("byFields is blank when given no fields")
    fun byFieldsBlankIfEmpty() =
        assertEquals("", Where.byFields(listOf()))

    @Test
    @DisplayName("byFields generates one numeric field | PostgreSQL")
    fun byFieldsOneFieldPostgres() {
        ForceDialect.postgres()
        assertEquals("(data->>'it')::numeric = :that", Where.byFields(listOf(Field.equal("it", 9, ":that"))))
    }

    @Test
    @DisplayName("byFields generates one alphanumeric field | PostgreSQL")
    fun byFieldsOneAlphaFieldPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'it' = :that", Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("byFields generates one field | SQLite")
    fun byFieldsOneFieldSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'it' = :that", Where.byFields(listOf(Field.equal("it", "", ":that"))))
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ default match | PostgreSQL")
    fun byFieldsMultipleDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "data->>'1' = :one AND (data->>'2')::numeric = :two AND data->>'3' = :three",
            Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))
            )
        )
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ default match | SQLite")
    fun byFieldsMultipleDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "data->>'1' = :one AND data->>'2' = :two AND data->>'3' = :three",
            Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three"))
            )
        )
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ ANY match | PostgreSQL")
    fun byFieldsMultipleAnyPostgres() {
        ForceDialect.postgres()
        assertEquals(
            "data->>'1' = :one OR (data->>'2')::numeric = :two OR data->>'3' = :three",
            Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("byFields generates multiple fields w/ ANY match | SQLite")
    fun byFieldsMultipleAnySQLite() {
        ForceDialect.sqlite()
        assertEquals(
            "data->>'1' = :one OR data->>'2' = :two OR data->>'3' = :three",
            Where.byFields(
                listOf(Field.equal("1", "", ":one"), Field.equal("2", 0L, ":two"), Field.equal("3", "", ":three")),
                FieldMatch.ANY
            )
        )
    }

    @Test
    @DisplayName("byId generates defaults for alphanumeric key | PostgreSQL")
    fun byIdDefaultAlphaPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'id' = :id", Where.byId(docId = ""))
    }

    @Test
    @DisplayName("byId generates defaults for numeric key | PostgreSQL")
    fun byIdDefaultNumericPostgres() {
        ForceDialect.postgres()
        assertEquals("(data->>'id')::numeric = :id", Where.byId(docId = 5))
    }

    @Test
    @DisplayName("byId generates defaults | SQLite")
    fun byIdDefaultSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'id' = :id", Where.byId(docId = ""))
    }

    @Test
    @DisplayName("byId generates named ID | PostgreSQL")
    fun byIdDefaultNamedPostgres() {
        ForceDialect.postgres()
        assertEquals("data->>'id' = :key", Where.byId<String>(":key"))
    }

    @Test
    @DisplayName("byId generates named ID | SQLite")
    fun byIdDefaultNamedSQLite() {
        ForceDialect.sqlite()
        assertEquals("data->>'id' = :key", Where.byId<String>(":key"))
    }

    @Test
    @DisplayName("jsonContains generates defaults | PostgreSQL")
    fun jsonContainsDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals("data @> :criteria", Where.jsonContains())
    }

    @Test
    @DisplayName("jsonContains generates named parameter | PostgreSQL")
    fun jsonContainsNamedPostgres() {
        ForceDialect.postgres()
        assertEquals("data @> :it", Where.jsonContains(":it"))
    }

    @Test
    @DisplayName("jsonContains fails | SQLite")
    fun jsonContainsFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { Where.jsonContains() }
    }

    @Test
    @DisplayName("jsonPathMatches generates defaults | PostgreSQL")
    fun jsonPathMatchDefaultPostgres() {
        ForceDialect.postgres()
        assertEquals("jsonb_path_exists(data, :path::jsonpath)", Where.jsonPathMatches())
    }

    @Test
    @DisplayName("jsonPathMatches generates named parameter | PostgreSQL")
    fun jsonPathMatchNamedPostgres() {
        ForceDialect.postgres()
        assertEquals("jsonb_path_exists(data, :jp::jsonpath)", Where.jsonPathMatches(":jp"))
    }

    @Test
    @DisplayName("jsonPathMatches fails | SQLite")
    fun jsonPathFailsSQLite() {
        ForceDialect.sqlite()
        assertThrows<DocumentException> { Where.jsonPathMatches() }
    }
}
