package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.AutoId
import solutions.bitbadger.documents.DocumentException

import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `AutoId` enum
 */
@DisplayName('Groovy | AutoId')
class AutoIdTest {

    @Test
    @DisplayName('Generates a UUID string')
    void generateUUID() {
        assertEquals 32, AutoId.generateUUID().length(), 'The UUID should have been a 32-character string'
    }

    @Test
    @DisplayName('Generates a random hex character string of an even length')
    void generateRandomStringEven() {
        def result = AutoId.generateRandomString 8
        assertEquals 8, result.length(), "There should have been 8 characters in $result"
    }

    @Test
    @DisplayName('Generates a random hex character string of an odd length')
    void generateRandomStringOdd() {
        def result = AutoId.generateRandomString 11
        assertEquals 11, result.length(), "There should have been 11 characters in $result"
    }

    @Test
    @DisplayName('Generates different random hex character strings')
    void generateRandomStringIsRandom() {
        def result1 = AutoId.generateRandomString 16
        def result2 = AutoId.generateRandomString 16
        assertNotEquals result1, result2, 'There should have been 2 different strings generated'
    }

    @Test
    @DisplayName('needsAutoId fails for null document')
    void needsAutoIdFailsForNullDocument() {
        assertThrows(DocumentException) { AutoId.needsAutoId(AutoId.DISABLED, null, 'id') }
    }

    @Test
    @DisplayName('needsAutoId fails for missing ID property')
    void needsAutoIdFailsForMissingId() {
        assertThrows(DocumentException) { AutoId.needsAutoId(AutoId.UUID, new IntIdClass(0), 'Id') }
    }

    @Test
    @DisplayName('needsAutoId returns false if disabled')
    void needsAutoIdFalseIfDisabled() {
        assertFalse AutoId.needsAutoId(AutoId.DISABLED, '', ''), 'Disabled Auto ID should always return false'
    }

    @Test
    @DisplayName('needsAutoId returns true for Number strategy and byte ID of 0')
    void needsAutoIdTrueForByteWithZero() {
        assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new ByteIdClass((byte) 0), 'id'),
                'Number Auto ID with 0 should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for Number strategy and byte ID of non-0')
    void needsAutoIdFalseForByteWithNonZero() {
        assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new ByteIdClass((byte) 77), 'id'),
                'Number Auto ID with 77 should return false')
    }

    @Test
    @DisplayName('needsAutoId returns true for Number strategy and short ID of 0')
    void needsAutoIdTrueForShortWithZero() {
        assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new ShortIdClass((short) 0), 'id'),
                'Number Auto ID with 0 should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for Number strategy and short ID of non-0')
    void needsAutoIdFalseForShortWithNonZero() {
        assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new ShortIdClass((short) 31), 'id'),
                'Number Auto ID with 31 should return false')
    }

    @Test
    @DisplayName('needsAutoId returns true for Number strategy and int ID of 0')
    void needsAutoIdTrueForIntWithZero() {
        assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new IntIdClass(0), 'id'),
                'Number Auto ID with 0 should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for Number strategy and int ID of non-0')
    void needsAutoIdFalseForIntWithNonZero() {
        assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new IntIdClass(6), 'id'),
                'Number Auto ID with 6 should return false')
    }

    @Test
    @DisplayName('needsAutoId returns true for Number strategy and long ID of 0')
    void needsAutoIdTrueForLongWithZero() {
        assertTrue(AutoId.needsAutoId(AutoId.NUMBER, new LongIdClass(0L), 'id'),
                'Number Auto ID with 0 should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for Number strategy and long ID of non-0')
    void needsAutoIdFalseForLongWithNonZero() {
        assertFalse(AutoId.needsAutoId(AutoId.NUMBER, new LongIdClass(2L), 'id'),
                'Number Auto ID with 2 should return false')
    }

    @Test
    @DisplayName('needsAutoId fails for Number strategy and non-number ID')
    void needsAutoIdFailsForNumberWithStringId() {
        assertThrows(DocumentException) { AutoId.needsAutoId(AutoId.NUMBER, new StringIdClass(''), 'id') }
    }

    @Test
    @DisplayName('needsAutoId returns true for UUID strategy and blank ID')
    void needsAutoIdTrueForUUIDWithBlank() {
        assertTrue(AutoId.needsAutoId(AutoId.UUID, new StringIdClass(''), 'id'),
                'UUID Auto ID with blank should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for UUID strategy and non-blank ID')
    void needsAutoIdFalseForUUIDNotBlank() {
        assertFalse(AutoId.needsAutoId(AutoId.UUID, new StringIdClass('howdy'), 'id'),
                'UUID Auto ID with non-blank should return false')
    }

    @Test
    @DisplayName('needsAutoId fails for UUID strategy and non-string ID')
    void needsAutoIdFailsForUUIDNonString() {
        assertThrows(DocumentException) { AutoId.needsAutoId(AutoId.UUID, new IntIdClass(5), 'id') }
    }

    @Test
    @DisplayName('needsAutoId returns true for Random String strategy and blank ID')
    void needsAutoIdTrueForRandomWithBlank() {
        assertTrue(AutoId.needsAutoId(AutoId.RANDOM_STRING, new StringIdClass(''), 'id'),
                'Random String Auto ID with blank should return true')
    }

    @Test
    @DisplayName('needsAutoId returns false for Random String strategy and non-blank ID')
    void needsAutoIdFalseForRandomNotBlank() {
        assertFalse(AutoId.needsAutoId(AutoId.RANDOM_STRING, new StringIdClass('full'), 'id'),
                'Random String Auto ID with non-blank should return false')
    }

    @Test
    @DisplayName('needsAutoId fails for Random String strategy and non-string ID')
    void needsAutoIdFailsForRandomNonString() {
        assertThrows(DocumentException) { AutoId.needsAutoId(AutoId.RANDOM_STRING, new ShortIdClass((short) 55), 'id') }
    }
}
