package solutions.bitbadger.documents.groovy.tests

import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.Parameter
import solutions.bitbadger.documents.ParameterType
import solutions.bitbadger.documents.query.RemoveFieldsQuery

import static Types.TEST_TABLE
import static org.junit.jupiter.api.Assertions.*

/**
 * Unit tests for the `RemoveFields` object
 */
@DisplayName('Groovy | Query | RemoveFieldsQuery')
class RemoveFieldsQueryTest {

    /**
     * Reset the dialect
     */
    @AfterEach
    void cleanUp() {
        ForceDialect.none()
    }

    @Test
    @DisplayName('byId generates correctly | PostgreSQL')
    void byIdPostgres() {
        ForceDialect.postgres()
        assertEquals("UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data->>'id' = :id".toString(),
                RemoveFieldsQuery.byId(TEST_TABLE, List.of(new Parameter(':name', ParameterType.STRING, '{a,z}'))),
                'Remove Fields query not constructed correctly')
    }

    @Test
    @DisplayName('byId generates correctly | SQLite')
    void byIdSQLite() {
        ForceDialect.sqlite()
        assertEquals(
                "UPDATE $TEST_TABLE SET data = json_remove(data, :name0, :name1) WHERE data->>'id' = :id".toString(),
                RemoveFieldsQuery.byId(TEST_TABLE, List.of(new Parameter(':name0', ParameterType.STRING, 'a'),
                        new Parameter(':name1', ParameterType.STRING, 'z'))),
                'Remove Field query not constructed correctly')
    }

    @Test
    @DisplayName('byFields generates correctly | PostgreSQL')
    void byFieldsPostgres() {
        ForceDialect.postgres()
        assertEquals("UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data->>'f' > :g".toString(),
                RemoveFieldsQuery.byFields(TEST_TABLE, List.of(new Parameter(':name', ParameterType.STRING, '{b,c}')),
                        List.of(Field.greater('f', '', ':g'))),
                'Remove Field query not constructed correctly')
    }

    @Test
    @DisplayName('byFields generates correctly | SQLite')
    void byFieldsSQLite() {
        ForceDialect.sqlite()
        assertEquals("UPDATE $TEST_TABLE SET data = json_remove(data, :name0, :name1) WHERE data->>'f' > :g".toString(),
                RemoveFieldsQuery.byFields(TEST_TABLE,
                        List.of(new Parameter(':name0', ParameterType.STRING, 'b'),
                                new Parameter(':name1', ParameterType.STRING, 'c')),
                        List.of(Field.greater('f', '', ':g'))),
                'Remove Field query not constructed correctly')
    }

    @Test
    @DisplayName('byContains generates correctly | PostgreSQL')
    void byContainsPostgres() {
        ForceDialect.postgres()
        assertEquals("UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE data @> :criteria".toString(),
                RemoveFieldsQuery.byContains(TEST_TABLE,
                        List.of(new Parameter(':name', ParameterType.STRING, '{m,n}'))),
                'Remove Field query not constructed correctly')
    }

    @Test
    @DisplayName('byContains fails | SQLite')
    void byContainsSQLite() {
        ForceDialect.sqlite()
        assertThrows(DocumentException) { RemoveFieldsQuery.byContains(TEST_TABLE, List.of()) }
    }

    @Test
    @DisplayName('byJsonPath generates correctly | PostgreSQL')
    void byJsonPathPostgres() {
        ForceDialect.postgres()
        assertEquals(
                "UPDATE $TEST_TABLE SET data = data - :name::text[] WHERE jsonb_path_exists(data, :path::jsonpath)"
                        .toString(),
                RemoveFieldsQuery.byJsonPath(TEST_TABLE,
                        List.of(new Parameter(':name', ParameterType.STRING, '{o,p}'))),
                'Remove Field query not constructed correctly')
    }

    @Test
    @DisplayName('byJsonPath fails | SQLite')
    void byJsonPathSQLite() {
        ForceDialect.sqlite()
        assertThrows(DocumentException) { RemoveFieldsQuery.byJsonPath(TEST_TABLE, List.of()) }
    }
}
