package solutions.bitbadger.documents.groovy.tests.integration

import solutions.bitbadger.documents.AutoId
import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.DocumentException
import solutions.bitbadger.documents.Field

import static org.junit.jupiter.api.Assertions.*
import static solutions.bitbadger.documents.groovy.tests.Types.TEST_TABLE

final class DocumentFunctions {

    static void insertDefault(ThrowawayDatabase db) {
        assertEquals 0L, db.conn.countAll(TEST_TABLE), 'There should be no documents in the table'
        def doc = new JsonDocument('turkey', 'yum', 5, new SubDocument('gobble', 'gobble!'))
        db.conn.insert TEST_TABLE, doc
        def after = db.conn.findAll TEST_TABLE, JsonDocument
        assertEquals 1, after.size(), 'There should be one document in the table'
        assertEquals doc.id, after[0].id, 'The document ID was not inserted correctly'
        assertEquals doc.value, after[0].value, 'The document value was not inserted correctly'
        assertEquals doc.numValue, after[0].numValue, 'The document numValue was not inserted correctly'
        assertNotNull doc.sub, "The subdocument was not inserted"
        assertEquals doc.sub.foo, after[0].sub.foo, 'The subdocument "foo" property was not inserted correctly'
        assertEquals doc.sub.bar, after[0].sub.bar, 'The subdocument "bar" property was not inserted correctly'
    }

    static void insertDupe(ThrowawayDatabase db) {
        db.conn.insert TEST_TABLE, new JsonDocument('a', '', 0, null)
        assertThrows(DocumentException, () -> db.conn.insert(TEST_TABLE, new JsonDocument('a', 'b', 22, null)),
                'Inserting a document with a duplicate key should have thrown an exception')
    }

    static void insertNumAutoId(ThrowawayDatabase db) {
        try {
            Configuration.autoIdStrategy = AutoId.NUMBER
            Configuration.idField = 'key'
            assertEquals 0L, db.conn.countAll(TEST_TABLE), 'There should be no documents in the table'

            db.conn.insert TEST_TABLE, new NumIdDocument(0, 'one')
            db.conn.insert TEST_TABLE, new NumIdDocument(0, 'two')
            db.conn.insert TEST_TABLE, new NumIdDocument(77, 'three')
            db.conn.insert TEST_TABLE, new NumIdDocument(0, 'four')

            def after = db.conn.findAll TEST_TABLE, NumIdDocument, List.of(Field.named('key'))
            assertEquals 4, after.size(), 'There should have been 4 documents returned'
            assertEquals('1|2|77|78',
                    after*.key*.toString().inject('') { acc, it -> acc == '' ? it : "$acc|$it" }.toString(),
                    'The IDs were not generated correctly')
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED
            Configuration.idField = 'id'
        }
    }

    static void insertUUIDAutoId(ThrowawayDatabase db) {
        try {
            Configuration.autoIdStrategy = AutoId.UUID
            assertEquals 0L, db.conn.countAll(TEST_TABLE), 'There should be no documents in the table'

            db.conn.insert TEST_TABLE, new JsonDocument('')

            def after = db.conn.findAll TEST_TABLE, JsonDocument
            assertEquals 1, after.size(), 'There should have been 1 document returned'
            assertEquals 32, after[0].id.length(), 'The ID was not generated correctly'
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED
        }
    }

    static void insertStringAutoId(ThrowawayDatabase db) {
        try {
            Configuration.autoIdStrategy = AutoId.RANDOM_STRING
            assertEquals 0L, db.conn.countAll(TEST_TABLE), 'There should be no documents in the table'

            db.conn.insert TEST_TABLE, new JsonDocument('')

            Configuration.idStringLength = 21
            db.conn.insert TEST_TABLE, new JsonDocument('')

            def after = db.conn.findAll TEST_TABLE, JsonDocument
            assertEquals 2, after.size(), 'There should have been 2 documents returned'
            assertEquals 16, after[0].id.length(), "The first document's ID was not generated correctly"
            assertEquals 21, after[1].id.length(), "The second document's ID was not generated correctly"
        } finally {
            Configuration.autoIdStrategy = AutoId.DISABLED
            Configuration.idStringLength = 16
        }
    }

    static void saveMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.save TEST_TABLE, new JsonDocument('two', '', 44)
        def tryDoc = db.conn.findById TEST_TABLE, 'two', JsonDocument
        assertTrue tryDoc.isPresent(), 'There should have been a document returned'
        def doc = tryDoc.get()
        assertEquals 'two', doc.id, 'An incorrect document was returned'
        assertEquals '', doc.value, 'The "value" field was not updated'
        assertEquals 44, doc.numValue, 'The "numValue" field was not updated'
        assertNull doc.sub, 'The "sub" field was not updated'
    }

    static void saveNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.save TEST_TABLE, new JsonDocument('test', '', 0, new SubDocument('a', 'b'))
        assertTrue(db.conn.findById(TEST_TABLE, 'test', JsonDocument).isPresent(),
                'The test document should have been saved')
    }

    static void updateMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        db.conn.update TEST_TABLE, 'one', new JsonDocument('one', 'howdy', 8, new SubDocument('y', 'z'))
        def tryDoc = db.conn.findById TEST_TABLE, 'one', JsonDocument
        assertTrue tryDoc.isPresent(), 'There should have been a document returned'
        def doc = tryDoc.get()
        assertEquals 'one', doc.id, 'An incorrect document was returned'
        assertEquals 'howdy', doc.value, 'The "value" field was not updated'
        assertEquals 8, doc.numValue, 'The "numValue" field was not updated'
        assertNotNull doc.sub, 'The sub-document should not be null'
        assertEquals 'y', doc.sub.foo, 'The sub-document "foo" field was not updated'
        assertEquals 'z', doc.sub.bar, 'The sub-document "bar" field was not updated'
    }

    static void updateNoMatch(ThrowawayDatabase db) {
        JsonDocument.load db
        assertFalse db.conn.existsById(TEST_TABLE, 'two-hundred')
        db.conn.update TEST_TABLE, 'two-hundred', new JsonDocument('two-hundred', '', 200)
        assertFalse db.conn.existsById(TEST_TABLE, 'two-hundred')
    }
}
