package solutions.bitbadger.documents.kotlinx

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.java.Exists as JvmExists
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.query.ExistsQuery
import java.sql.Connection

/**
 * Functions to determine whether documents exist
 */
object Exists {

    /**
     * Determine a document's existence by its ID
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param docId The ID of the document to be checked
     * @param conn The connection on which the existence check should be executed
     * @return True if the document exists, false if not
     */
    fun <TKey> byId(tableName: String, docId: TKey, conn: Connection) =
        JvmExists.byId(tableName, docId, conn)

    /**
     * Determine a document's existence by its ID
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param docId The ID of the document to be checked
     * @return True if the document exists, false if not
     */
    fun <TKey> byId(tableName: String, docId: TKey) =
        JvmExists.byId(tableName, docId)

    /**
     * Determine document existence using a field comparison
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param conn The connection on which the existence check should be executed
     * @return True if any matching documents exist, false if not
     */
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null, conn: Connection) =
        JvmExists.byFields(tableName, fields, howMatched, conn)

    /**
     * Determine document existence using a field comparison
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param fields The fields which should be compared
     * @param howMatched How the fields should be matched
     * @return True if any matching documents exist, false if not
     */
    fun byFields(tableName: String, fields: Collection<Field<*>>, howMatched: FieldMatch? = null) =
        JvmExists.byFields(tableName, fields, howMatched)

    /**
     * Determine document existence using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param criteria The object for which JSON containment should be checked
     * @param conn The connection on which the existence check should be executed
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains, conn: Connection) =
        conn.customScalar(
            ExistsQuery.byContains(tableName),
            listOf(Parameters.json(":criteria", criteria)),
            Results::toExists
        )

    /**
     * Determine document existence using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param criteria The object for which JSON containment should be checked
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    inline fun <reified TContains> byContains(tableName: String, criteria: TContains) =
        Configuration.dbConn().use { byContains(tableName, criteria, it) }

    /**
     * Determine document existence using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param path The JSON path comparison to match
     * @param conn The connection on which the existence check should be executed
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String, conn: Connection) =
        JvmExists.byJsonPath(tableName, path, conn)

    /**
     * Determine document existence using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param path The JSON path comparison to match
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    fun byJsonPath(tableName: String, path: String) =
        JvmExists.byJsonPath(tableName, path)
}
