package solutions.bitbadger.documents.kotlinx.tests.integration

import solutions.bitbadger.documents.*
import solutions.bitbadger.documents.kotlinx.extensions.*
import solutions.bitbadger.documents.kotlinx.tests.ArrayDocument
import solutions.bitbadger.documents.kotlinx.tests.JsonDocument
import solutions.bitbadger.documents.kotlinx.tests.NumIdDocument
import solutions.bitbadger.documents.kotlinx.tests.TEST_TABLE
import kotlin.test.assertEquals
import kotlin.test.assertNotNull
import kotlin.test.assertNull
import kotlin.test.assertTrue

/**
 * Integration tests for the `Find` object
 */
object FindFunctions {

    fun allDefault(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(5, db.conn.findAll<JsonDocument>(TEST_TABLE).size, "There should have been 5 documents returned")
    }

    fun allAscending(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findAll<JsonDocument>(TEST_TABLE, listOf(Field.named("id")))
        assertEquals(5, docs.size, "There should have been 5 documents returned")
        assertEquals(
            "five|four|one|three|two",
            docs.joinToString("|") { it.id },
            "The documents were not ordered correctly"
        )
    }

    fun allDescending(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findAll<JsonDocument>(TEST_TABLE, listOf(Field.named("id DESC")))
        assertEquals(5, docs.size, "There should have been 5 documents returned")
        assertEquals(
            "two|three|one|four|five",
            docs.joinToString("|") { it.id },
            "The documents were not ordered correctly"
        )
    }

    fun allNumOrder(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findAll<JsonDocument>(
            TEST_TABLE,
            listOf(Field.named("sub.foo NULLS LAST"), Field.named("n:numValue"))
        )
        assertEquals(5, docs.size, "There should have been 5 documents returned")
        assertEquals(
            "two|four|one|three|five",
            docs.joinToString("|") { it.id },
            "The documents were not ordered correctly"
        )
    }

    fun allEmpty(db: ThrowawayDatabase) =
        assertEquals(0, db.conn.findAll<JsonDocument>(TEST_TABLE).size, "There should have been no documents returned")

    fun byIdString(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findById<String, JsonDocument>(TEST_TABLE, "two")
        assertNotNull(doc, "The document should have been returned")
        assertEquals("two", doc.id, "An incorrect document was returned")
    }

    fun byIdNumber(db: ThrowawayDatabase) {
        Configuration.idField = "key"
        try {
            db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
            val doc = db.conn.findById<Int, NumIdDocument>(TEST_TABLE, 18)
            assertNotNull(doc, "The document should have been returned")
        } finally {
            Configuration.idField = "id"
        }
    }

    fun byIdNotFound(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertNull(
            db.conn.findById<String, JsonDocument>(TEST_TABLE, "x"),
            "There should have been no document returned"
        )
    }

    fun byFieldsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByFields<JsonDocument>(
            TEST_TABLE,
            listOf(Field.any("value", listOf("blue", "purple")), Field.exists("sub")),
            FieldMatch.ALL
        )
        assertEquals(1, docs.size, "There should have been a document returned")
        assertEquals("four", docs[0].id, "The incorrect document was returned")
    }

    fun byFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByFields<JsonDocument>(
            TEST_TABLE,
            listOf(Field.equal("value", "purple")),
            orderBy = listOf(Field.named("id"))
        )
        assertEquals(2, docs.size, "There should have been 2 documents returned")
        assertEquals("five|four", docs.joinToString("|") { it.id }, "The documents were not ordered correctly")
    }

    fun byFieldsMatchNumIn(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByFields<JsonDocument>(TEST_TABLE, listOf(Field.any("numValue", listOf(2, 4, 6, 8))))
        assertEquals(1, docs.size, "There should have been a document returned")
        assertEquals("three", docs[0].id, "The incorrect document was returned")
    }

    fun byFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            0,
            db.conn.findByFields<JsonDocument>(TEST_TABLE, listOf(Field.greater("numValue", 100))).size,
            "There should have been no documents returned"
        )
    }

    fun byFieldsMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        val docs =
            db.conn.findByFields<ArrayDocument>(TEST_TABLE, listOf(Field.inArray("values", TEST_TABLE, listOf("c"))))
        assertEquals(2, docs.size, "There should have been two documents returned")
        assertTrue(listOf("first", "second").contains(docs[0].id), "An incorrect document was returned (${docs[0].id})")
        assertTrue(listOf("first", "second").contains(docs[1].id), "An incorrect document was returned (${docs[1].id})")
    }

    fun byFieldsNoMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        assertEquals(
            0,
            db.conn.findByFields<ArrayDocument>(
                TEST_TABLE,
                listOf(Field.inArray("values", TEST_TABLE, listOf("j")))
            ).size,
            "There should have been no documents returned"
        )
    }

    fun byContainsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByContains<JsonDocument, Map<String, String>>(TEST_TABLE, mapOf("value" to "purple"))
        assertEquals(2, docs.size, "There should have been 2 documents returned")
        assertTrue(listOf("four", "five").contains(docs[0].id), "An incorrect document was returned (${docs[0].id})")
        assertTrue(listOf("four", "five").contains(docs[1].id), "An incorrect document was returned (${docs[1].id})")
    }

    fun byContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByContains<JsonDocument, Map<String, Map<String, String>>>(
            TEST_TABLE,
            mapOf("sub" to mapOf("foo" to "green")),
            listOf(Field.named("value"))
        )
        assertEquals(2, docs.size, "There should have been 2 documents returned")
        assertEquals("two|four", docs.joinToString("|") { it.id }, "The documents were not ordered correctly")
    }

    fun byContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            0,
            db.conn.findByContains<JsonDocument, Map<String, String>>(TEST_TABLE, mapOf("value" to "indigo")).size,
            "There should have been no documents returned"
        )
    }

    fun byJsonPathMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ > 10)")
        assertEquals(2, docs.size, "There should have been 2 documents returned")
        assertTrue(listOf("four", "five").contains(docs[0].id), "An incorrect document was returned (${docs[0].id})")
        assertTrue(listOf("four", "five").contains(docs[1].id), "An incorrect document was returned (${docs[1].id})")
    }

    fun byJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val docs = db.conn.findByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ > 10)", listOf(Field.named("id")))
        assertEquals(2, docs.size, "There should have been 2 documents returned")
        assertEquals("five|four", docs.joinToString("|") { it.id }, "The documents were not ordered correctly")
    }

    fun byJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            0,
            db.conn.findByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ > 100)").size,
            "There should have been no documents returned"
        )
    }

    fun firstByFieldsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByFields<JsonDocument>(TEST_TABLE, listOf(Field.equal("value", "another")))
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("two", doc.id, "The incorrect document was returned")
    }

    fun firstByFieldsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByFields<JsonDocument>(TEST_TABLE, listOf(Field.equal("sub.foo", "green")))
        assertNotNull(doc, "There should have been a document returned")
        assertTrue(listOf("two", "four").contains(doc.id), "An incorrect document was returned (${doc.id})")
    }

    fun firstByFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByFields<JsonDocument>(
            TEST_TABLE,
            listOf(Field.equal("sub.foo", "green")),
            orderBy = listOf(Field.named("n:numValue DESC"))
        )
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("four", doc.id, "An incorrect document was returned")
    }

    fun firstByFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertNull(
            db.conn.findFirstByFields(TEST_TABLE, listOf(Field.equal("value", "absent"))),
            "There should have been no document returned"
        )
    }

    fun firstByContainsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByContains<JsonDocument, Map<String, String>>(TEST_TABLE, mapOf("value" to "FIRST!"))
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("one", doc.id, "An incorrect document was returned")
    }

    fun firstByContainsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByContains<JsonDocument, Map<String, String>>(TEST_TABLE, mapOf("value" to "purple"))
        assertNotNull(doc, "There should have been a document returned")
        assertTrue(listOf("four", "five").contains(doc.id), "An incorrect document was returned (${doc.id})")
    }

    fun firstByContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByContains<JsonDocument, Map<String, String>>(
            TEST_TABLE,
            mapOf("value" to "purple"),
            listOf(Field.named("sub.bar NULLS FIRST"))
        )
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("five", doc.id, "An incorrect document was returned")
    }

    fun firstByContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertNull(
            db.conn.findFirstByContains<JsonDocument, Map<String, String>>(
                TEST_TABLE,
                mapOf("value" to "indigo")
            ), "There should have been no document returned"
        )
    }

    fun firstByJsonPathMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ == 10)")
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("two", doc.id, "An incorrect document was returned")
    }

    fun firstByJsonPathMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ > 10)")
        assertNotNull(doc, "There should have been a document returned")
        assertTrue(listOf("four", "five").contains(doc.id), "An incorrect document was returned (${doc.id})")
    }

    fun firstByJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val doc = db.conn.findFirstByJsonPath<JsonDocument>(
            TEST_TABLE,
            "$.numValue ? (@ > 10)",
            listOf(Field.named("id DESC"))
        )
        assertNotNull(doc, "There should have been a document returned")
        assertEquals("four", doc.id, "An incorrect document was returned")
    }

    fun firstByJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertNull(
            db.conn.findFirstByJsonPath<JsonDocument>(TEST_TABLE, "$.numValue ? (@ > 100)"),
            "There should have been no document returned"
        )
    }
}
