package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{Field, FieldMatch}
import solutions.bitbadger.documents.java.Exists as CoreExists

import java.sql.Connection
import _root_.scala.jdk.CollectionConverters.*

/**
 * Functions to determine whether documents exist
 */
object Exists:

  /**
   * Determine a document's existence by its ID
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param docId     The ID of the document to be checked
   * @param conn      The connection on which the existence check should be executed
   * @return True if the document exists, false if not
   * @throws DocumentException If no dialect has been configured
   */
  def byId[Key](tableName: String, docId: Key, conn: Connection): Boolean =
    CoreExists.byId(tableName, docId, conn)

  /**
   * Determine a document's existence by its ID (creates connection)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param docId     The ID of the document to be checked
   * @return True if the document exists, false if not
   * @throws DocumentException If no connection string has been set
   */
  def byId[Key](tableName: String, docId: Key): Boolean =
    CoreExists.byId(tableName, docId)

  /**
   * Determine document existence using a field comparison
   *
   * @param tableName  The name of the table in which document existence should be checked
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @param conn       The connection on which the existence check should be executed
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If no dialect has been configured, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch], conn: Connection): Boolean =
    CoreExists.byFields(tableName, fields.asJava, howMatched.orNull, conn)

  /**
   * Determine document existence using a field comparison (creates connection)
   *
   * @param tableName  The name of the table in which document existence should be checked
   * @param fields     The fields which should be compared
   * @param howMatched How the fields should be matched
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If no connection string has been set, or if parameters are invalid
   */
  def byFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Boolean =
    CoreExists.byFields(tableName, fields.asJava, howMatched.orNull)

  /**
   * Determine document existence using a JSON containment query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param criteria  The object for which JSON containment should be checked
   * @param conn      The connection on which the existence check should be executed
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A, conn: Connection): Boolean =
    CoreExists.byContains(tableName, criteria, conn)

  /**
   * Determine document existence using a JSON containment query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param criteria  The object for which JSON containment should be checked
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byContains[A](tableName: String, criteria: A): Boolean =
    CoreExists.byContains(tableName, criteria)

  /**
   * Determine document existence using a JSON Path match query (PostgreSQL only)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param path      The JSON path comparison to match
   * @param conn      The connection on which the existence check should be executed
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String, conn: Connection): Boolean =
    CoreExists.byJsonPath(tableName, path, conn)

  /**
   * Determine document existence using a JSON Path match query (PostgreSQL only; creates connection)
   *
   * @param tableName The name of the table in which document existence should be checked
   * @param path      The JSON path comparison to match
   * @return True if any matching documents exist, false if not
   * @throws DocumentException If no connection string has been set, or if called on a SQLite connection
   */
  def byJsonPath(tableName: String, path: String): Boolean =
    CoreExists.byJsonPath(tableName, path)
