package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{DisplayName, Test}
import solutions.bitbadger.documents.{AutoId, DocumentException}

@DisplayName("Scala | AutoId")
class AutoIdTest:

  @Test
  @DisplayName("Generates a UUID string")
  def generateUUID(): Unit =
    assertEquals(32, AutoId.generateUUID().length(), "The UUID should have been a 32-character string")

  @Test
  @DisplayName("Generates a random hex character string of an even length")
  def generateRandomStringEven(): Unit =
    val result = AutoId.generateRandomString(8)
    assertEquals(8, result.length(), s"There should have been 8 characters in $result")

  @Test
  @DisplayName("Generates a random hex character string of an odd length")
  def generateRandomStringOdd(): Unit =
    val result = AutoId.generateRandomString(11)
    assertEquals(11, result.length(), s"There should have been 11 characters in $result")

  @Test
  @DisplayName("Generates different random hex character strings")
  def generateRandomStringIsRandom(): Unit =
    val result1 = AutoId.generateRandomString(16)
    val result2 = AutoId.generateRandomString(16)
    assertNotEquals(result1, result2, "There should have been 2 different strings generated")

  @Test
  @DisplayName("needsAutoId fails for null document")
  def needsAutoIdFailsForNullDocument(): Unit =
    assertThrows(classOf[DocumentException], () => AutoId.needsAutoId(AutoId.DISABLED, null, "id"))

  @Test
  @DisplayName("needsAutoId fails for missing ID property")
  def needsAutoIdFailsForMissingId(): Unit =
    assertThrows(classOf[DocumentException], () => AutoId.needsAutoId(AutoId.UUID, IntIdClass(0), "Id"))


  @Test
  @DisplayName("needsAutoId returns false if disabled")
  def needsAutoIdFalseIfDisabled(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.DISABLED, "", ""), "Disabled Auto ID should always return false")

  @Test
  @DisplayName("needsAutoId returns true for Number strategy and byte ID of 0")
  def needsAutoIdTrueForByteWithZero(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.NUMBER, ByteIdClass(0), "id"), "Number Auto ID with 0 should return true")

  @Test
  @DisplayName("needsAutoId returns false for Number strategy and byte ID of non-0")
  def needsAutoIdFalseForByteWithNonZero(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.NUMBER, ByteIdClass(77), "id"), "Number Auto ID with 77 should return false")

  @Test
  @DisplayName("needsAutoId returns true for Number strategy and short ID of 0")
  def needsAutoIdTrueForShortWithZero(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.NUMBER, ShortIdClass(0), "id"), "Number Auto ID with 0 should return true")

  @Test
  @DisplayName("needsAutoId returns false for Number strategy and short ID of non-0")
  def needsAutoIdFalseForShortWithNonZero(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.NUMBER, ShortIdClass(31), "id"), "Number Auto ID with 31 should return false")

  @Test
  @DisplayName("needsAutoId returns true for Number strategy and int ID of 0")
  def needsAutoIdTrueForIntWithZero(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.NUMBER, IntIdClass(0), "id"), "Number Auto ID with 0 should return true")

  @Test
  @DisplayName("needsAutoId returns false for Number strategy and int ID of non-0")
  def needsAutoIdFalseForIntWithNonZero(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.NUMBER, IntIdClass(6), "id"), "Number Auto ID with 6 should return false")

  @Test
  @DisplayName("needsAutoId returns true for Number strategy and long ID of 0")
  def needsAutoIdTrueForLongWithZero(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.NUMBER, LongIdClass(0), "id"), "Number Auto ID with 0 should return true")

  @Test
  @DisplayName("needsAutoId returns false for Number strategy and long ID of non-0")
  def needsAutoIdFalseForLongWithNonZero(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.NUMBER, LongIdClass(2), "id"), "Number Auto ID with 2 should return false")

  @Test
  @DisplayName("needsAutoId fails for Number strategy and non-number ID")
  def needsAutoIdFailsForNumberWithStringId(): Unit =
    assertThrows(classOf[DocumentException], () => AutoId.needsAutoId(AutoId.NUMBER, StringIdClass(""), "id"))

  @Test
  @DisplayName("needsAutoId returns true for UUID strategy and blank ID")
  def needsAutoIdTrueForUUIDWithBlank(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.UUID, StringIdClass(""), "id"), "UUID Auto ID with blank should return true")

  @Test
  @DisplayName("needsAutoId returns false for UUID strategy and non-blank ID")
  def needsAutoIdFalseForUUIDNotBlank(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.UUID, StringIdClass("howdy"), "id"),
      "UUID Auto ID with non-blank should return false")

  @Test
  @DisplayName("needsAutoId fails for UUID strategy and non-string ID")
  def needsAutoIdFailsForUUIDNonString(): Unit =
    assertThrows(classOf[DocumentException], () => AutoId.needsAutoId(AutoId.UUID, IntIdClass(5), "id"))

  @Test
  @DisplayName("needsAutoId returns true for Random String strategy and blank ID")
  def needsAutoIdTrueForRandomWithBlank(): Unit =
    assertTrue(AutoId.needsAutoId(AutoId.RANDOM_STRING, StringIdClass(""), "id"),
      "Random String Auto ID with blank should return true")

  @Test
  @DisplayName("needsAutoId returns false for Random String strategy and non-blank ID")
  def needsAutoIdFalseForRandomNotBlank(): Unit =
    assertFalse(AutoId.needsAutoId(AutoId.RANDOM_STRING, StringIdClass("full"), "id"),
      "Random String Auto ID with non-blank should return false")

  @Test
  @DisplayName("needsAutoId fails for Random String strategy and non-string ID")
  def needsAutoIdFailsForRandomNonString(): Unit =
    assertThrows(classOf[DocumentException], () => AutoId.needsAutoId(AutoId.RANDOM_STRING, ShortIdClass(55), "id"))
