package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{AfterEach, DisplayName, Test}
import solutions.bitbadger.documents.{Dialect, DocumentException, DocumentIndex}
import solutions.bitbadger.documents.query.DefinitionQuery

import scala.jdk.CollectionConverters.*

@DisplayName("Scala | Query | DefinitionQuery")
class DefinitionQueryTest:

  /**
   * Clear the connection string (resets Dialect)
   */
  @AfterEach
  def cleanUp(): Unit =
    ForceDialect.none()

  @Test
  @DisplayName("ensureTableFor generates correctly")
  def ensureTableFor(): Unit =
    assertEquals("CREATE TABLE IF NOT EXISTS my.table (data JSONB NOT NULL)",
      DefinitionQuery.ensureTableFor("my.table", "JSONB"), "CREATE TABLE statement not constructed correctly")

  @Test
  @DisplayName("ensureTable generates correctly | PostgreSQL")
  def ensureTablePostgres(): Unit =
    ForceDialect.postgres()
    assertEquals(s"CREATE TABLE IF NOT EXISTS $TEST_TABLE (data JSONB NOT NULL)",
      DefinitionQuery.ensureTable(TEST_TABLE))

  @Test
  @DisplayName("ensureTable generates correctly | SQLite")
  def ensureTableSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals(s"CREATE TABLE IF NOT EXISTS $TEST_TABLE (data TEXT NOT NULL)",
      DefinitionQuery.ensureTable(TEST_TABLE))

  @Test
  @DisplayName("ensureTable fails when no dialect is set")
  def ensureTableFailsUnknown(): Unit =
    assertThrows(classOf[DocumentException], () => DefinitionQuery.ensureTable(TEST_TABLE))

  @Test
  @DisplayName("ensureKey generates correctly with schema")
  def ensureKeyWithSchema(): Unit =
    assertEquals("CREATE UNIQUE INDEX IF NOT EXISTS idx_table_key ON test.table ((data->>'id'))",
      DefinitionQuery.ensureKey("test.table", Dialect.POSTGRESQL),
      "CREATE INDEX for key statement with schema not constructed correctly")

  @Test
  @DisplayName("ensureKey generates correctly without schema")
  def ensureKeyWithoutSchema(): Unit =
    assertEquals(s"CREATE UNIQUE INDEX IF NOT EXISTS idx_${TEST_TABLE}_key ON $TEST_TABLE ((data->>'id'))",
      DefinitionQuery.ensureKey(TEST_TABLE, Dialect.SQLITE),
      "CREATE INDEX for key statement without schema not constructed correctly")

  @Test
  @DisplayName("ensureIndexOn generates multiple fields and directions")
  def ensureIndexOnMultipleFields(): Unit =
    assertEquals("CREATE INDEX IF NOT EXISTS idx_table_gibberish ON test.table " +
        "((data->>'taco'), (data->>'guac') DESC, (data->>'salsa') ASC)",
      DefinitionQuery.ensureIndexOn("test.table", "gibberish", List("taco", "guac DESC", "salsa ASC").asJava,
        Dialect.POSTGRESQL),
      "CREATE INDEX for multiple field statement not constructed correctly")

  @Test
  @DisplayName("ensureIndexOn generates nested field | PostgreSQL")
  def ensureIndexOnNestedPostgres(): Unit =
    assertEquals(s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data#>>'{a,b,c}'))",
      DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List("a.b.c").asJava, Dialect.POSTGRESQL),
      "CREATE INDEX for nested PostgreSQL field incorrect")

  @Test
  @DisplayName("ensureIndexOn generates nested field | SQLite")
  def ensureIndexOnNestedSQLite(): Unit =
    assertEquals(s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_nest ON $TEST_TABLE ((data->'a'->'b'->>'c'))",
      DefinitionQuery.ensureIndexOn(TEST_TABLE, "nest", List("a.b.c").asJava, Dialect.SQLITE),
      "CREATE INDEX for nested SQLite field incorrect")

  @Test
  @DisplayName("ensureDocumentIndexOn generates Full | PostgreSQL")
  def ensureDocumentIndexOnFullPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals(s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data)",
      DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL),
      "CREATE INDEX for full document index incorrect")

  @Test
  @DisplayName("ensureDocumentIndexOn generates Optimized | PostgreSQL")
  def ensureDocumentIndexOnOptimizedPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals(
      s"CREATE INDEX IF NOT EXISTS idx_${TEST_TABLE}_document ON $TEST_TABLE USING GIN (data jsonb_path_ops)",
      DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.OPTIMIZED),
      "CREATE INDEX for optimized document index incorrect")

  @Test
  @DisplayName("ensureDocumentIndexOn fails | SQLite")
  def ensureDocumentIndexOnFailsSQLite(): Unit =
    ForceDialect.sqlite()
    assertThrows(classOf[DocumentException],
      () => DefinitionQuery.ensureDocumentIndexOn(TEST_TABLE, DocumentIndex.FULL))
