package solutions.bitbadger.documents.scala.tests

import org.junit.jupiter.api.Assertions.*
import org.junit.jupiter.api.{AfterEach, DisplayName, Test}
import solutions.bitbadger.documents.{AutoId, Configuration, DocumentException}
import solutions.bitbadger.documents.query.DocumentQuery

@DisplayName("Scala | Query | DocumentQuery")
class DocumentQueryTest:

  /**
   * Clear the connection string (resets Dialect)
   */
  @AfterEach
  def cleanUp(): Unit =
    ForceDialect.none()

  @Test
  @DisplayName("insert generates no auto ID | PostgreSQL")
  def insertNoAutoPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals(s"INSERT INTO $TEST_TABLE VALUES (:data)", DocumentQuery.insert(TEST_TABLE))

  @Test
  @DisplayName("insert generates no auto ID | SQLite")
  def insertNoAutoSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals(s"INSERT INTO $TEST_TABLE VALUES (:data)", DocumentQuery.insert(TEST_TABLE))

  @Test
  @DisplayName("insert generates auto number | PostgreSQL")
  def insertAutoNumberPostgres(): Unit =
    ForceDialect.postgres()
    assertEquals(s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || ('{\"id\":' " +
        s"|| (SELECT COALESCE(MAX((data->>'id')::numeric), 0) + 1 FROM $TEST_TABLE) || '}')::jsonb)",
      DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER))

  @Test
  @DisplayName("insert generates auto number | SQLite")
  def insertAutoNumberSQLite(): Unit =
    ForceDialect.sqlite()
    assertEquals(s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', " +
        s"(SELECT coalesce(max(data->>'id'), 0) + 1 FROM $TEST_TABLE)))",
      DocumentQuery.insert(TEST_TABLE, AutoId.NUMBER))

  @Test
  @DisplayName("insert generates auto UUID | PostgreSQL")
  def insertAutoUUIDPostgres(): Unit =
    ForceDialect.postgres()
    val query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID)
    assertTrue(query.startsWith(s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\""),
      s"Query start not correct (actual: $query)")
    assertTrue(query.endsWith("\"}')"), "Query end not correct")

  @Test
  @DisplayName("insert generates auto UUID | SQLite")
  def insertAutoUUIDSQLite(): Unit =
    ForceDialect.sqlite()
    val query = DocumentQuery.insert(TEST_TABLE, AutoId.UUID)
    assertTrue(query.startsWith(s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '"),
      s"Query start not correct (actual: $query)")
    assertTrue(query.endsWith("'))"), "Query end not correct")

  @Test
  @DisplayName("insert generates auto random string | PostgreSQL")
  def insertAutoRandomPostgres(): Unit =
    try
      ForceDialect.postgres()
      Configuration.idStringLength = 8
      val query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING)
      assertTrue(query.startsWith(s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\""),
        s"Query start not correct (actual: $query)")
      assertTrue(query.endsWith("\"}')"), "Query end not correct")
      assertEquals(8, query.replace(s"INSERT INTO $TEST_TABLE VALUES (:data::jsonb || '{\"id\":\"", "")
          .replace("\"}')", "").length,
        "Random string length incorrect")
    finally
      Configuration.idStringLength = 16

  @Test
  @DisplayName("insert generates auto random string | SQLite")
  def insertAutoRandomSQLite(): Unit =
    ForceDialect.sqlite()
    val query = DocumentQuery.insert(TEST_TABLE, AutoId.RANDOM_STRING)
    assertTrue(query.startsWith(s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '"),
      s"Query start not correct (actual: $query)")
    assertTrue(query.endsWith("'))"), "Query end not correct")
    assertEquals(Configuration.idStringLength,
      query.replace(s"INSERT INTO $TEST_TABLE VALUES (json_set(:data, '$$.id', '", "").replace("'))", "").length,
      "Random string length incorrect")

  @Test
  @DisplayName("insert fails when no dialect is set")
  def insertFailsUnknown(): Unit =
    assertThrows(classOf[DocumentException], () => DocumentQuery.insert(TEST_TABLE))

  @Test
  @DisplayName("save generates correctly")
  def save(): Unit =
    ForceDialect.postgres()
    assertEquals(
      s"INSERT INTO $TEST_TABLE VALUES (:data) ON CONFLICT ((data->>'id')) DO UPDATE SET data = EXCLUDED.data",
      DocumentQuery.save(TEST_TABLE), "INSERT ON CONFLICT UPDATE statement not constructed correctly")

  @Test
  @DisplayName("update generates successfully")
  def update(): Unit =
    assertEquals(s"UPDATE $TEST_TABLE SET data = :data", DocumentQuery.update(TEST_TABLE),
      "Update query not constructed correctly")
