package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.{Configuration, Field}
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE

import scala.jdk.CollectionConverters.*

object FindFunctions:

  /** Generate IDs as a pipe-delimited string */
  private def docIds(docs: List[JsonDocument]) =
    docs.map(_.id).reduce((ids, docId) => s"$ids|$docId")

  def allDefault(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(5, db.conn.findAll[JsonDocument](TEST_TABLE).size, "There should have been 5 documents returned")

  def allAscending(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findAll[JsonDocument](TEST_TABLE, Field.named("id") :: Nil)
    assertEquals(5, docs.size, "There should have been 5 documents returned")
    assertEquals("five|four|one|three|two", docIds(docs), "The documents were not ordered correctly")

  def allDescending(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findAll[JsonDocument](TEST_TABLE, Field.named("id DESC") :: Nil)
    assertEquals(5, docs.size, "There should have been 5 documents returned")
    assertEquals("two|three|one|four|five", docIds(docs), "The documents were not ordered correctly")

  def allNumOrder(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findAll[JsonDocument](TEST_TABLE,
      Field.named("sub.foo NULLS LAST") :: Field.named("n:numValue") :: Nil)
    assertEquals(5, docs.size, "There should have been 5 documents returned")
    assertEquals("two|four|one|three|five", docIds(docs), "The documents were not ordered correctly")

  def allEmpty(db: ThrowawayDatabase): Unit =
    assertEquals(0, db.conn.findAll[JsonDocument](TEST_TABLE).size, "There should have been no documents returned")

  def byIdString(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findById[String, JsonDocument](TEST_TABLE, "two")
    assertTrue(doc.isDefined, "The document should have been returned")
    assertEquals("two", doc.get.id, "An incorrect document was returned")

  def byIdNumber(db: ThrowawayDatabase): Unit =
    Configuration.idField = "key"
    try
      db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
      val doc = db.conn.findById[Int, NumIdDocument](TEST_TABLE, 18)
      assertTrue(doc.isDefined, "The document should have been returned")
    finally
      Configuration.idField = "id"

  def byIdNotFound(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.findById[String, JsonDocument](TEST_TABLE, "x").isDefined,
      "There should have been no document returned")

  def byFieldsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByFields[JsonDocument](TEST_TABLE,
      Field.any("value", ("blue" :: "purple" :: Nil).asJava) :: Nil, orderBy = Field.exists("sub") :: Nil)
    assertEquals(1, docs.size, "There should have been a document returned")
    assertEquals("four", docs.head.id, "The incorrect document was returned")

  def byFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByFields[JsonDocument](TEST_TABLE, Field.equal("value", "purple") :: Nil,
      orderBy = Field.named("id") :: Nil)
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    assertEquals("five|four", docIds(docs), "The documents were not ordered correctly")

  def byFieldsMatchNumIn(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByFields[JsonDocument](TEST_TABLE,
      Field.any("numValue", (2 :: 4 :: 6 :: 8 :: Nil).asJava) :: Nil)
    assertEquals(1, docs.size, "There should have been a document returned")
    assertEquals("three", docs.head.id, "The incorrect document was returned")

  def byFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(0, db.conn.findByFields[JsonDocument](TEST_TABLE, Field.greater("numValue", 100) :: Nil).size,
      "There should have been no documents returned")

  def byFieldsMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    val docs = db.conn.findByFields[ArrayDocument](TEST_TABLE,
      Field.inArray("values", TEST_TABLE, ("c" :: Nil).asJava) :: Nil)
    assertEquals(2, docs.size, "There should have been two documents returned")
    assertTrue(("first" :: "second" :: Nil).contains(docs.head.id),
      s"An incorrect document was returned (${docs.head.id}")
    assertTrue(("first" :: "second" :: Nil).contains(docs(1).id), s"An incorrect document was returned (${docs(1).id})")

  def byFieldsNoMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    assertEquals(0,
      db.conn.findByFields[ArrayDocument](TEST_TABLE,
        Field.inArray("values", TEST_TABLE, ("j" :: Nil).asJava) :: Nil).size,
      "There should have been no documents returned")

  def byContainsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE, Map.Map1("value", "purple"))
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    assertTrue(("four" :: "five" :: Nil).contains(docs.head.id),
      s"An incorrect document was returned (${docs.head.id})")
    assertTrue(("four" :: "five" :: Nil).contains(docs(1).id), s"An incorrect document was returned (${docs(1).id})")

  def byContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByContains[JsonDocument, Map.Map1[String, Map.Map1[String, String]]](TEST_TABLE,
      Map.Map1("sub", Map.Map1("foo", "green")), Field.named("value") :: Nil)
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    assertEquals("two|four", docIds(docs), "The documents were not ordered correctly")

  def byContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(0,
      db.conn.findByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE, Map.Map1("value", "indigo")).size,
      "There should have been no documents returned")

  def byJsonPathMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 10)")
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    assertTrue(("four" :: "five" :: Nil).contains(docs.head.id),
      s"An incorrect document was returned (${docs.head.id})")
    assertTrue(("four" :: "five" :: Nil).contains(docs(1).id), s"An incorrect document was returned (${docs(1).id})")

  def byJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val docs = db.conn.findByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 10)", Field.named("id") :: Nil)
    assertEquals(2, docs.size, "There should have been 2 documents returned")
    assertEquals("five|four", docIds(docs), "The documents were not ordered correctly")

  def byJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals(0, db.conn.findByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 100)").size,
      "There should have been no documents returned")

  def firstByFieldsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByFields[JsonDocument](TEST_TABLE, Field.equal("value", "another") :: Nil)
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("two", doc.get.id, "The incorrect document was returned")

  def firstByFieldsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByFields[JsonDocument](TEST_TABLE, Field.equal("sub.foo", "green") :: Nil)
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertTrue(("two" :: "four" :: Nil).contains(doc.get.id), s"An incorrect document was returned (${doc.get.id})")

  def firstByFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByFields[JsonDocument](TEST_TABLE, Field.equal("sub.foo", "green") :: Nil,
      orderBy = Field.named("n:numValue DESC") :: Nil)
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("four", doc.get.id, "An incorrect document was returned")

  def firstByFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.findFirstByFields[JsonDocument](TEST_TABLE, Field.equal("value", "absent") :: Nil).isDefined,
      "There should have been no document returned")

  def firstByContainsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE,
      Map.Map1("value", "FIRST!"))
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("one", doc.get.id, "An incorrect document was returned")

  def firstByContainsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE,
      Map.Map1("value", "purple"))
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertTrue(("four" :: "five" :: Nil).contains(doc.get.id), s"An incorrect document was returned (${doc.get.id})")

  def firstByContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE,
      Map.Map1("value", "purple"), Field.named("sub.bar NULLS FIRST") :: Nil)
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("five", doc.get.id, "An incorrect document was returned")

  def firstByContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.findFirstByContains[JsonDocument, Map.Map1[String, String]](TEST_TABLE,
      Map.Map1("value", "indigo")).isDefined, "There should have been no document returned")

  def firstByJsonPathMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ == 10)")
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("two", doc.get.id, "An incorrect document was returned")

  def firstByJsonPathMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 10)")
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertTrue(("four" :: "five" :: Nil).contains(doc.get.id), s"An incorrect document was returned (${doc.get.id})")

  def firstByJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val doc = db.conn.findFirstByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 10)",
      Field.named("id DESC") :: Nil)
    assertTrue(doc.isDefined, "There should have been a document returned")
    assertEquals("four", doc.get.id, "An incorrect document was returned")

  def firstByJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertFalse(db.conn.findFirstByJsonPath[JsonDocument](TEST_TABLE, "$.numValue ? (@ > 100)").isDefined,
      "There should have been no document returned")
