package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.{Configuration, Dialect, Field, FieldMatch}
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE

import java.io.{PrintWriter, StringWriter}
import scala.jdk.CollectionConverters.*

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
object JsonFunctions:

  /**
   * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
   * This function will do a crude string replacement to match the target string based on the dialect being tested.
   *
   * @param json The JSON which should be returned
   * @return The actual expected JSON based on the database being tested
   */
  def maybeJsonB(json: String): String =
    Configuration.dialect() match
      case Dialect.SQLITE => json
      case Dialect.POSTGRESQL => json.replace("\":", "\": ").replace(",\"", ", \"")

  /**
   * Create a snippet of JSON to find a document ID
   *
   * @param id The ID of the document
   * @return A connection-aware ID to check for presence and positioning
   */
  private def docId(id: String): String =
    maybeJsonB(s"""{"id":"$id"""")

  private def checkAllDefault(json: String): Unit =
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.one), s"Document 'one' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.two), s"Document 'two' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.three), s"Document 'three' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found in JSON ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("one")), s"Document 'one' not found in JSON ($json)")
        assertTrue(json.contains(docId("two")), s"Document 'two' not found in JSON ($json)")
        assertTrue(json.contains(docId("three")), s"Document 'three' not found in JSON ($json)")
        assertTrue(json.contains(docId("four")), s"Document 'four' not found in JSON ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found in JSON ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def allDefault(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkAllDefault(db.conn.jsonAll(TEST_TABLE))

  def writeAllDefault(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonAll(TEST_TABLE, writer)
    checkAllDefault(output.toString)

  private def checkAllEmpty(json: String): Unit =
    assertEquals("[]", json, "There should have been no documents returned")

  def allEmpty(db: ThrowawayDatabase): Unit =
    checkAllEmpty(db.conn.jsonAll(TEST_TABLE))

  def writeAllEmpty(db: ThrowawayDatabase): Unit =
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonAll(TEST_TABLE, writer)
    checkAllEmpty(output.toString)

  private def checkByIdString(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"An incorrect document was returned ($json)")

  def byIdString(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByIdString(db.conn.jsonById(TEST_TABLE, "two"))

  def writeByIdString(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonById(TEST_TABLE, writer, "two")
    checkByIdString(output.toString)

  private def checkByIdNumber(json: String): Unit =
    assertEquals(maybeJsonB("""{"key":18,"text":"howdy"}"""), json, "The document should have been found by numeric ID")

  def byIdNumber(db: ThrowawayDatabase): Unit =
    Configuration.idField = "key"
    try
      db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
      checkByIdNumber(db.conn.jsonById(TEST_TABLE, 18))
    finally
      Configuration.idField = "id"

  def writeByIdNumber(db: ThrowawayDatabase): Unit =
    Configuration.idField = "key"
    try
      db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
      val output = StringWriter()
      val writer = PrintWriter(output)
      db.conn.writeJsonById(TEST_TABLE, writer, 18)
      checkByIdNumber(output.toString)
    finally
      Configuration.idField = "id"

  private def checkByIdNotFound(json: String): Unit =
    assertEquals("{}", json, "There should have been no document returned")

  def byIdNotFound(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByIdNotFound(db.conn.jsonById(TEST_TABLE, "x"))

  def writeByIdNotFound(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonById(TEST_TABLE, writer, "x")
    checkByIdNotFound(output.toString)

  private def checkByFieldsMatch(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.four}]", json, "The incorrect document was returned")
      case Dialect.POSTGRESQL =>
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("four")), s"The incorrect document was returned ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByFieldsMatch(db.conn.jsonByFields(TEST_TABLE,
      Field.any("value", ("blue" :: "purple" :: Nil).asJava) :: Field.exists("sub") :: Nil, Some(FieldMatch.ALL)))

  def writeByFieldsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer,
      Field.any("value", ("blue" :: "purple" :: Nil).asJava) :: Field.exists("sub") :: Nil, Some(FieldMatch.ALL))
    checkByFieldsMatch(output.toString)

  private def checkByFieldsMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val fiveIdx = json.indexOf(docId("five"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(fiveIdx >= 0, s"Document 'five' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(fiveIdx < fourIdx, s"Document 'five' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByFieldsMatchOrdered(db.conn.jsonByFields(TEST_TABLE, Field.equal("value", "purple") :: Nil, None,
      Field.named("id") :: Nil))

  def writeByFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer, Field.equal("value", "purple") :: Nil, None, Field.named("id") :: Nil)
    checkByFieldsMatchOrdered(output.toString)

  private def checkByFieldsMatchNumIn(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.three}]", json, "The incorrect document was returned")
      case Dialect.POSTGRESQL =>
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("three")), s"The incorrect document was returned ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatchNumIn(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByFieldsMatchNumIn(db.conn.jsonByFields(TEST_TABLE,
      Field.any("numValue", (2 :: 4 :: 6 :: 8 :: Nil).asJava) :: Nil))

  def writeByFieldsMatchNumIn(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer, Field.any("numValue", (2 :: 4 :: 6 :: 8 :: Nil).asJava) :: Nil)
    checkByFieldsMatchNumIn(output.toString)

  private def checkByFieldsNoMatch(json: String): Unit =
    assertEquals("[]", json, "There should have been no documents returned")

  def byFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByFieldsNoMatch(db.conn.jsonByFields(TEST_TABLE, Field.greater("numValue", 100) :: Nil))

  def writeByFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer, Field.greater("numValue", 100) :: Nil)
    checkByFieldsNoMatch(output.toString)

  private def checkByFieldsMatchInArray(json: String): Unit =
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    assertTrue(json.contains(docId("first")), s"The 'first' document was not found ($json)")
    assertTrue(json.contains(docId("second")), s"The 'second' document was not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    checkByFieldsMatchInArray(db.conn.jsonByFields(TEST_TABLE,
      Field.inArray("values", TEST_TABLE, ("c" :: Nil).asJava) :: Nil))

  def writeByFieldsMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer, Field.inArray("values", TEST_TABLE, ("c" :: Nil).asJava) :: Nil)
    checkByFieldsMatchInArray(output.toString)

  private def checkByFieldsNoMatchInArray(json: String): Unit =
    assertEquals("[]", json, "There should have been no documents returned")

  def byFieldsNoMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    checkByFieldsNoMatchInArray(db.conn.jsonByFields(TEST_TABLE,
      Field.inArray("values", TEST_TABLE, ("j" :: Nil).asJava) :: Nil))

  def writeByFieldsNoMatchInArray(db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByFields(TEST_TABLE, writer, Field.inArray("values", TEST_TABLE, ("j" :: Nil).asJava) :: Nil)
    checkByFieldsNoMatchInArray(output.toString)

  private def checkByContainsMatch(json: String): Unit =
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")), s"Document 'four' not found ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byContainsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByContainsMatch(db.conn.jsonByContains(TEST_TABLE, Map.Map1("value", "purple")))

  def writeByContainsMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByContains(TEST_TABLE, writer, Map.Map1("value", "purple"))
    checkByContainsMatch(output.toString)

  private def checkByContainsMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.two},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val twoIdx = json.indexOf(docId("two"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(twoIdx >= 0, s"Document 'two' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(twoIdx < fourIdx, s"Document 'two' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByContainsMatchOrdered(db.conn.jsonByContains(TEST_TABLE, Map.Map1("sub", Map.Map1("foo", "green")),
      Field.named("value") :: Nil))

  def writeByContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByContains(TEST_TABLE, writer, Map.Map1("sub", Map.Map1("foo", "green")),
      Field.named("value") :: Nil)
    checkByContainsMatchOrdered(output.toString)

  private def checkByContainsNoMatch(json: String): Unit =
    assertEquals("[]", json, "There should have been no documents returned")

  def byContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByContainsNoMatch(db.conn.jsonByContains(TEST_TABLE, Map.Map1("value", "indigo")))

  def writeByContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByContains(TEST_TABLE, writer, Map.Map1("value", "indigo"))
    checkByContainsNoMatch(output.toString)

  private def checkByJsonPathMatch(json: String): Unit =
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")), s"Document 'four' not found ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byJsonPathMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByJsonPathMatch(db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)"))

  def writeByJsonPathMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)")
    checkByJsonPathMatch(output.toString)

  private def checkByJsonPathMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val fiveIdx = json.indexOf(docId("five"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(fiveIdx >= 0, s"Document 'five' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(fiveIdx < fourIdx, s"Document 'five' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByJsonPathMatchOrdered(db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", Field.named("id") :: Nil))

  def writeByJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)", Field.named("id") :: Nil)
    checkByJsonPathMatchOrdered(output.toString)

  private def checkByJsonPathNoMatch(json: String): Unit =
    assertEquals("[]", json, "There should have been no documents returned")

  def byJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkByJsonPathNoMatch(db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"))

  def writeByJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 100)")
    checkByJsonPathNoMatch(output.toString)

  private def checkFirstByFieldsMatchOne(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "The incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"The incorrect document was returned ($json)")

  def firstByFieldsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByFieldsMatchOne(db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("value", "another") :: Nil))

  def writeFirstByFieldsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByFields(TEST_TABLE, writer, Field.equal("value", "another") :: Nil)
    checkFirstByFieldsMatchOne(output.toString)

  private def checkFirstByFieldsMatchMany(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
          s"Expected document 'two' or 'four' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("two")) || json.contains(docId("four")),
          s"Expected document 'two' or 'four' ($json)")

  def firstByFieldsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByFieldsMatchMany(db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("sub.foo", "green") :: Nil))

  def writeFirstByFieldsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByFields(TEST_TABLE, writer, Field.equal("sub.foo", "green") :: Nil)
    checkFirstByFieldsMatchMany(output.toString)

  private def checkFirstByFieldsMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.four, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("four")), s"An incorrect document was returned ($json)")

  def firstByFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByFieldsMatchOrdered(db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("sub.foo", "green") :: Nil, None,
      Field.named("n:numValue DESC") :: Nil))

  def writeFirstByFieldsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByFields(TEST_TABLE, writer, Field.equal("sub.foo", "green") :: Nil, None,
      Field.named("n:numValue DESC") :: Nil)

  private def checkFirstByFieldsNoMatch(json: String): Unit =
    assertEquals("{}", json, "There should have been no document returned")

  def firstByFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByFieldsNoMatch(db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("value", "absent") :: Nil))

  def writeFirstByFieldsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByFields(TEST_TABLE, writer, Field.equal("value", "absent") :: Nil)
    checkFirstByFieldsNoMatch(output.toString)

  private def checkFirstByContainsMatchOne(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.one, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("one")), s"An incorrect document was returned ($json)")

  def firstByContainsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByContainsMatchOne(db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "FIRST!")))

  def writeFirstByContainsMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.Map1("value", "FIRST!"))
    checkFirstByContainsMatchOne(output.toString)

  private def checkFirstByContainsMatchMany(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
          s"Expected document 'four' or 'five' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
          s"Expected document 'four' or 'five' ($json)")

  def firstByContainsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByContainsMatchMany(db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "purple")))

  def writeFirstByContainsMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.Map1("value", "purple"))
    checkFirstByContainsMatchMany(output.toString)

  private def checkFirstByContainsMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.five, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("five")), s"An incorrect document was returned ($json)")

  def firstByContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByContainsMatchOrdered(db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "purple"),
      Field.named("sub.bar NULLS FIRST") :: Nil))

  def writeFirstByContainsMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.Map1("value", "purple"),
      Field.named("sub.bar NULLS FIRST") :: Nil)
    checkFirstByContainsMatchOrdered(output.toString)

  private def checkFirstByContainsNoMatch(json: String): Unit =
    assertEquals("{}", json, "There should have been no document returned")

  def firstByContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByContainsNoMatch(db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "indigo")))

  def writeFirstByContainsNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByContains(TEST_TABLE, writer, Map.Map1("value", "indigo"))
    checkFirstByContainsNoMatch(output.toString)

  private def checkFirstByJsonPathMatchOne(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"An incorrect document was returned ($json)")

  def firstByJsonPathMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByJsonPathMatchOne(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ == 10)"))

  def writeFirstByJsonPathMatchOne(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ == 10)")
    checkFirstByJsonPathMatchOne(output.toString)

  private def checkFirstByJsonPathMatchMany(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
          s"Expected document 'four' or 'five' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
          s"Expected document 'four' or 'five' ($json)")

  def firstByJsonPathMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByJsonPathMatchMany(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)"))

  def writeFirstByJsonPathMatchMany(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)")
    checkFirstByJsonPathMatchMany(output.toString)

  private def checkFirstByJsonPathMatchOrdered(json: String): Unit =
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.four, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("four")), s"An incorrect document was returned ($json)")

  def firstByJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByJsonPathMatchOrdered(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)",
      Field.named("id DESC") :: Nil))

  def writeFirstByJsonPathMatchOrdered(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 10)", Field.named("id DESC") :: Nil)
    checkFirstByJsonPathMatchOrdered(output.toString)

  private def checkFirstByJsonPathNoMatch(json: String): Unit =
    assertEquals("{}", json, "There should have been no document returned")

  def firstByJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    checkFirstByJsonPathNoMatch(db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"))

  def writeFirstByJsonPathNoMatch(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val output = StringWriter()
    val writer = PrintWriter(output)
    db.conn.writeJsonFirstByJsonPath(TEST_TABLE, writer, "$.numValue ? (@ > 100)")
    checkFirstByJsonPathNoMatch(output.toString)
